from typing import Dict, Optional, Union

from volcenginesdkarkruntime.types.chat import (
    ChatCompletion,
    ChatCompletionChunk,
)

from ark.component.v3.llm.base import BaseChatLanguageModel, handle_function_call
from ark.core.idl.ark_protocol import (
    ArkChatCompletionChunk,
    ArkChatRequest,
    ArkChatResponse,
    CallableFunction,
    FunctionCallMode,
)
from ark.core.idl.common_protocol import ActionUsage


class ActionCall(BaseChatLanguageModel):
    async def handle_function_call(
        self,
        request: ArkChatRequest,
        response: Union[
            ChatCompletionChunk, ChatCompletion, ArkChatCompletionChunk, ArkChatResponse
        ],
        available_functions: Optional[Dict[str, CallableFunction]] = None,
        function_call_mode: Optional[FunctionCallMode] = FunctionCallMode.SEQUENTIAL,
    ) -> bool:
        is_more_request = await handle_function_call(
            request,
            response,
            available_functions,
            function_call_mode,
            include_details=True,
        )
        if (
            isinstance(response, ArkChatCompletionChunk)
            and response.bot_usage
            and response.bot_usage.action_details
        ):
            response.bot_usage.action_usage = [
                ActionUsage(count=1, name=detail.name)
                for detail in response.bot_usage.action_details
            ]
        return is_more_request



"common protocol for GPTEngine"

from __future__ import annotations

from enum import Enum
from typing import Any, Dict, List, Optional, TypeVar, Union

from pydantic import BaseModel, Field, field_validator
from volcenginesdkarkruntime.types.completion_usage import CompletionUsage


class Embedding(BaseModel):
    index: Optional[int] = None
    embedding: List[float] = Field(default_factory=list)
    object: Optional[str] = None


class UserInfo(BaseModel):
    accountId: str
    userId: str
    authorization_token: str = Field(repr=False, exclude=True)


class FunctionCall(BaseModel):
    name: Optional[str] = None
    arguments: Optional[str] = None


class Server(BaseModel):
    url: str
    description: Optional[str] = None


class Action(BaseModel):
    server: Server
    http_method: str = Field(default="POST")
    path: str
    tool: Tool


class Function(BaseModel):
    name: str
    description: Optional[str] = None
    parameters: Optional[Dict] = None
    examples: Optional[List[str]] = None


class Tool(BaseModel):
    type: str
    function: Optional[Function] = None
    options: Optional[Dict[str, Any]] = None


class ApiCall(BaseModel):
    name: str
    headers: Optional[Any] = None
    parameters: Optional[Any] = None
    json_datas: Optional[Any] = None


class CoverImage(BaseModel):
    url: str = ""
    width: int = 0
    height: int = 0


class Reference(BaseModel):
    url: Optional[str] = None
    idx: Optional[int] = None
    logo_url: Optional[str] = None
    mobile_url: Optional[str] = None
    site_name: Optional[str] = Field(default=None, description="来源")
    title: Optional[str] = Field(default=None, description="标题")
    summary: Optional[str] = None
    publish_time: Optional[str] = Field(default=None, description="发布时间")
    cover_image: Optional[CoverImage] = None
    freshness_info: Optional[str] = Field(default=None, description="时效性")
    extra: Optional[Dict[str, Any]] = None

    # deprecated
    pc_url: Optional[str] = None

    # knowledge base
    doc_id: Optional[str] = Field(default=None, description="文档ID")
    doc_name: Optional[str] = Field(default=None, description="文档名")
    doc_type: Optional[str] = Field(default=None, description="文档类型")
    doc_title: Optional[str] = Field(default=None, description="文档标题")
    chunk_title: Optional[str] = Field(default=None, description="分块标题")
    chunk_id: Optional[str] = Field(default=None, description="分块编号")
    collection_name: Optional[str] = None
    project: Optional[str] = None


class File(BaseModel):
    name: Optional[str] = None
    url: Optional[str] = None
    uri: Optional[str] = None


class Error(BaseModel):
    code: str
    code_n: Optional[int] = None
    message: str


class ArkError(BaseModel):
    code: str
    message: str
    param: Optional[str] = None
    type: Optional[str] = None


class Request(BaseModel):
    stream: bool = False


class Response(BaseModel):
    error: Optional[Union[Error, ArkError]] = None


class ModelDesc(BaseModel):
    name: Optional[str] = None
    version: Optional[str] = None
    endpoint_id: Optional[str] = None
    authorization_token: Optional[str] = None


class ChatMessageImageContent(BaseModel):
    url: Optional[str] = None
    detail: Optional[str] = None
    image_bytes: Optional[bytes] = None


class ChatMessageContent(BaseModel):
    type: Optional[str] = None
    text: Optional[str] = None
    image_url: Optional[ChatMessageImageContent] = None


class Parameters(BaseModel):
    max_prompt_tokens: Optional[int] = None
    max_new_tokens: Optional[int] = None
    min_new_tokens: Optional[int] = None
    max_tokens: Optional[int] = None

    temperature: Optional[float] = None
    top_p: Optional[float] = None
    top_k: Optional[int] = None

    presence_penalty: Optional[float] = None
    frequency_penalty: Optional[float] = None
    repetition_penalty: Optional[float] = None
    do_sample: Optional[bool] = None
    logprobs: Optional[int] = None

    stop: Optional[List[str]] = None
    logit_bias: Optional[Dict[int, float]] = None
    guidance: Optional[bool] = None

    def merge_from(self, other: Union[Dict, Parameters]) -> Parameters:
        other_dict = other if isinstance(other, Dict) else other.dict()
        merged_dict = self.dict()

        for key in merged_dict.keys():
            value = other_dict.get(key, None)
            if value is not None:
                merged_dict[key] = value

        return self.__class__(**merged_dict)

    def merge_to(self, other: Union[Dict, Parameters]) -> Parameters:
        self_dict = self.dict()
        merged_dict = other if isinstance(other, Dict) else other.dict()

        for key in self_dict.keys():
            value = self_dict.get(key, None)
            if value is not None:
                merged_dict[key] = value

        return self.__class__(**merged_dict)


class ChatRole(str, Enum):
    SYSTEM = "system"
    ASSISTANT = "assistant"
    USER = "user"
    FUNCTION = "function"
    TOOL = "tool"


class ChatUsage(BaseModel):
    name: Optional[str] = None
    prompt_tokens: int = 0
    completion_tokens: int = 0
    total_tokens: int = 0

    def __iadd__(self, others: Union[ChatUsage, List[ChatUsage]]) -> ChatUsage:
        if not isinstance(others, list):
            others = [others]

        for usage in others:
            self.prompt_tokens += usage.prompt_tokens
            self.completion_tokens += usage.completion_tokens
            self.total_tokens += usage.total_tokens

        return self

    def __add__(self, others: Union[ChatUsage, List[ChatUsage]]) -> ChatUsage:
        if not isinstance(others, list):
            others = [others]

        total_usage = ChatUsage(
            name=self.name,
            prompt_tokens=self.prompt_tokens,
            completion_tokens=self.completion_tokens,
            total_tokens=self.total_tokens,
        )
        for usage in others:
            total_usage.prompt_tokens += usage.prompt_tokens
            total_usage.completion_tokens += usage.completion_tokens
            total_usage.total_tokens += usage.total_tokens

        return total_usage


class ToolOutputType(str, Enum):
    TOOL = "tool"
    EXCEPTION = "exception"


class ToolOutput(BaseModel):
    type: ToolOutputType
    data: Optional[Union[Any, ExceptionInfo]] = None


class ExceptionInfo(BaseModel):
    type: Optional[str] = None
    message: Optional[str] = None


class ActionUsage(BaseModel):
    # deprecated
    name: Optional[str] = None
    prompt_tokens: Optional[int] = None
    completion_tokens: Optional[int] = None
    total_tokens: Optional[int] = None
    search_count: Optional[int] = None

    # latest
    action_name: Optional[str] = None
    """
    action name 
    """
    count: Optional[int] = None
    """
    count for calling the action
    """

    def __iadd__(
        self, others: Union[ActionUsage, List[ActionUsage]]
    ) -> List[ActionUsage]:
        if not isinstance(others, list):
            return [self, others]
        else:
            others.append(self)
            return others

    def __add__(
        self, others: Union[ActionUsage, List[ActionUsage]]
    ) -> List[ActionUsage]:
        if not isinstance(others, list):
            return [self, others]
        else:
            others.append(self)
            return others


class ActionDetails(BaseModel):
    name: str
    """
    action name, e.g. "WebBrowsingPlus"
    """
    count: int = 0
    """
    count for calling the action, e.g. 1
    """
    tool_details: List[ToolDetails] = Field(default_factory=list)
    """
    details about calling the tool
    """


class ToolDetails(BaseModel):
    name: str
    """
    tool name, e.g. "Search"
    """
    input: Any
    """
    input for calling the tool
    """
    output: Union[Any, ToolOutput]
    """
    output for calling the tool
    """
    created_at: Optional[int] = None
    """
    created time in milliseconds since the Epoch.
    """
    completed_at: Optional[int] = None
    """
    completed time in milliseconds since the Epoch.
    """


class BotUsage(BaseModel):
    model_usage: Optional[List[Union[ChatUsage, CompletionUsage]]] = Field(
        default_factory=list
    )
    action_usage: Optional[List[ActionUsage]] = Field(default_factory=list)
    action_details: Optional[List[ActionDetails]] = Field(default_factory=list)

    def __iadd__(self, others: Union[BotUsage, List[BotUsage]]) -> BotUsage:
        if not isinstance(others, list):
            others = [others]

        for usage in others:
            if self.model_usage and usage.model_usage:
                self.model_usage.extend(usage.model_usage)
            elif usage.model_usage:
                self.model_usage = usage.model_usage

            if self.action_usage and usage.action_usage:
                self.action_usage.extend(usage.action_usage)
            elif usage.action_usage:
                self.action_usage = usage.action_usage

            if self.action_details and usage.action_details:
                self.action_details.extend(usage.action_details)
            elif usage.action_details:
                self.action_details = usage.action_details

        return self

    def __add__(self, others: Union[BotUsage, List[BotUsage]]) -> BotUsage:
        if not isinstance(others, list):
            others = [others]

        total_usage = BotUsage(
            model_usage=self.model_usage or [],
            action_usage=self.action_usage or [],
            action_details=self.action_details or [],
        )
        for usage in others:
            if (
                usage.action_usage
                and total_usage.action_usage
                and len(usage.action_usage) > 0
            ):
                total_usage.action_usage.extend(usage.action_usage)
            if (
                usage.model_usage
                and total_usage.model_usage
                and len(usage.model_usage) > 0
            ):
                total_usage.model_usage.extend(usage.model_usage)
            if (
                usage.action_details
                and total_usage.action_details
                and len(usage.action_details) > 0
            ):
                total_usage.action_details.extend(usage.action_details)

        return total_usage


class Logprobs(BaseModel):
    text_offset: Optional[List[int]] = None
    token_logprobs: Optional[List[float]] = None
    tokens: Optional[List[str]] = None
    top_logprobs: Optional[List[Dict[str, Any]]] = None


class CertRequest(Request):
    model: Union[ModelDesc, None] = None


class CertResponse(Response):
    model: Union[ModelDesc, None] = None
    cert: str = ""


class ClassificationRequest(Request):
    model: Union[ModelDesc, None] = None
    req_id: Optional[str] = None
    query: Optional[str] = None
    labels: List[str] = Field(default_factory=list)


class ClassificationResponse(Response):
    label: Optional[str] = None
    label_logprobos: Optional[Dict[str, LabelLogprobosValue]] = None
    usage: Optional[ChatUsage] = None
    req_id: Optional[str] = None
    error: Optional[Error] = None


class TokenizationRequest(Request):
    model: Union[ModelDesc, None] = None
    req_id: Optional[str] = None
    text: Optional[str] = None


class TokenizationResponse(Response):
    total_tokens: Optional[int] = None
    tokens: Optional[List[str]] = None
    token_ids: Optional[List[int]] = None
    offset_mapping: Optional[List[List[int]]] = None
    req_id: Optional[str] = None
    error: Optional[Error] = None


class EmbeddingsRequest(Request):
    model: Union[ModelDesc, None] = None
    req_id: Optional[str] = None
    input: Optional[List[str]] = None
    encoding_format: Optional[str] = None
    user: Optional[str] = None


class EmbeddingsResponse(Response):
    usage: Optional[ChatUsage] = None
    req_id: Optional[str] = None
    error: Optional[Error] = None
    object: Optional[str] = None
    data: Optional[List[Embedding]] = None


class LabelLogprobosValue(BaseModel):
    tokens: List[str] = Field(default_factory=list)
    token_logprobs: List[float] = Field(default_factory=list)
    req_id: Optional[str] = None


class ChatMessage(BaseModel):
    role: Optional[ChatRole] = None
    content: str = Field(repr=False, default=None)
    multi_contents: Optional[List[ChatMessageContent]] = None
    name: Optional[str] = None
    function_call: Optional[FunctionCall] = None
    references: Optional[List[Reference]] = None
    reference: Optional[Reference] = None
    files: Optional[List[File]] = None


class ChatRequest(Request):
    model: ModelDesc
    messages: Optional[List[ChatMessage]] = None
    parameters: Optional[Parameters] = None
    stream: bool = False
    req_id: Optional[str] = None
    crypto_token: Optional[str] = None
    functions: Optional[List[Function]] = None
    tools: Optional[List[Tool]] = None
    plugins: Optional[List[str]] = None
    model_arch: Optional[str] = None

    extra: Optional[Dict[str, str]] = None

    def set_extra(self, key: str, value: str) -> None:
        if self.extra is None:
            self.extra = {}
        self.extra[key] = value

    @field_validator("model")
    def validate_model(cls, v: Optional[ModelDesc]) -> Optional[ModelDesc]:
        if not v:
            raise ValueError("Model is required")
        if not v.name and not v.endpoint_id:
            raise ValueError("Model name or endpoint_id is required")
        return v

    @field_validator("messages")
    def validate_messages(
        cls, v: Optional[List[ChatMessage]]
    ) -> Optional[List[ChatMessage]]:
        if not v:
            raise ValueError("Messages are required")

        pre_role = None  # user, assistant, function, system
        for msg in v:
            if not msg.role:
                raise ValueError("Role is required")
            if msg.content is None:
                msg.content = ""

            if msg.role == ChatRole.SYSTEM:
                continue

            if msg.role == ChatRole.USER:
                assert (
                    not pre_role or pre_role == ChatRole.ASSISTANT
                ), f"{pre_role} message should not followed by a user message"

            if msg.role == ChatRole.ASSISTANT:
                assert pre_role in [
                    ChatRole.USER,
                    ChatRole.FUNCTION,
                ], "Assistant message should follow a user or function message"

            if msg.role == ChatRole.FUNCTION:
                assert (
                    pre_role == ChatRole.ASSISTANT
                ), "Function message should follow an assistant message"

            pre_role = msg.role

        return v


class ChatChoice(BaseModel):
    index: Optional[int] = None
    message: ChatMessage
    finish_reason: Optional[str] = None
    logprobs: Optional[Logprobs] = None


class ChatResponse(Response):
    req_id: str = ""
    created: Optional[int] = None
    error: Optional[Error] = None
    choice: Optional[ChatChoice] = None
    choices: List[ChatChoice] = Field(default_factory=list)
    usage: Optional[ChatUsage] = None
    extra: Optional[Dict[str, str]] = None

    def set_extra(self, key: str, value: str) -> None:
        if self.extra is None:
            self.extra = dict()
        self.extra[key] = value

    @staticmethod
    def merge(responses: List[ChatResponse]) -> ChatResponse:
        merged = responses[0]
        for resp in responses[1:]:
            for i, j in zip(merged.choices, resp.choices):
                i.message.content += j.message.content
                i.finish_reason = j.finish_reason
        merged.usage = responses[-1].usage
        return merged


class JsonDType(str, Enum):
    Object = "object"
    Number = "number"
    Integer = "integer"
    String = "string"
    Array = "array"
    Null = "null"
    Boolean = "boolean"


class AudioMediaType(str, Enum):
    WAV = "wav"
    MP3 = "mp3"
    AAC = "aac"
    OGG_OPUS = "ogg_opus"
    PCM = "pcm"


class JsonDefinition(BaseModel):
    # Type specifies the data type of the schema.
    type: JsonDType
    # Description is the description of the schema.
    description: Optional[str] = None
    enum: Optional[List[str]] = None
    properties: Optional[Dict[str, JsonDefinition]] = None
    required: Optional[List[str]] = None
    items: Optional[JsonDefinition] = None
    property_order: Optional[List[str]] = None


class ToolCall(BaseModel):
    id: Optional[str] = None
    type: Optional[str] = None
    function: Optional[FunctionCall] = None


class Context(BaseModel):
    request_id: str = ""
    client_request_id: str = ""
    account_id: str = ""
    resource_id: str = ""
    resource_type: str = ""


RequestType = TypeVar("RequestType", bound=Request, contravariant=True)
ResponseType = TypeVar("ResponseType", bound=Response, covariant=True)



import asyncio
import time
import warnings
from typing import Any, Generic, Optional, TypeVar

import structlog

logger = structlog.stdlib.get_logger()
T = TypeVar("T")


class Singleton(Generic[T]):
    _instance: Optional[T] = None
    _lock: Optional[asyncio.Lock] = None

    @classmethod
    async def get_instance_async(cls, *args: Any, **kwargs: Any) -> T:
        if not cls._instance:
            if not cls._lock:
                cls._lock = asyncio.Lock()
            async with cls._lock:
                if not cls._instance:
                    self = cls(*args, **kwargs)
                    assert hasattr(
                        self, "async_init"
                    ), "async singletons must define async_init function"
                    await self.async_init()
                    cls._instance = self
                    logger.debug("singleton class initialized", name=cls.__name__)
        return cls._instance

    @classmethod
    def get_instance_sync(cls, *args: Any, **kwargs: Any) -> T:
        if not cls._instance:
            self = cls(*args, **kwargs)
            assert not hasattr(
                self, "async_init"
            ), f"class {cls.__name__} init with get_instance_async."
            cls._instance = self
            logger.debug("singleton class initialized", name=cls.__name__)
        return cls._instance


class LazyLoadSingleton(Generic[T]):
    _instance: Optional[T] = None
    _lock: Optional[asyncio.Lock] = None
    REFRESH_TIME_INTERVAL = 300
    _refresh_time: float = 0

    @classmethod
    def is_outdated(cls) -> bool:
        return time.time() - cls._refresh_time > cls.REFRESH_TIME_INTERVAL

    @classmethod
    async def get_instance_async(cls, *args: Any, **kwargs: Any) -> T:
        if not cls._instance or cls.is_outdated():
            if not cls._lock:
                cls._lock = asyncio.Lock()
            async with cls._lock:
                if (not cls._instance) or cls.is_outdated():
                    assert hasattr(
                        cls, "async_init"
                    ), "async singletons must define async_init function"
                    cls._instance = await cls.async_init(*args, **kwargs)
                    cls._refresh_time = time.time()
                    logger.debug("singleton class initialized", name=cls.__name__)
        return cls._instance

    @classmethod
    def get_instance_sync(cls, *args: Any, **kwargs: Any) -> T:
        if (not cls._instance) or cls.is_outdated():
            self = cls(*args, **kwargs)
            if hasattr(self, "async_init"):
                warnings.warn(
                    f"class {cls.__name__} has async_init function, "
                    f"init it with get_instance_async."
                )
            cls._instance = self
            cls._refresh_time = time.time()
            logger.debug("singleton class initialized", name=cls.__name__)
        return cls._instance



from __future__ import annotations

import json
import time
from datetime import datetime
from typing import Any, Dict, List, Optional, Tuple, Union
import pytz
import jinja2
from jinja2 import Template
from langchain.prompts.chat import BaseChatPromptTemplate
from langchain.schema.messages import (
    AIMessage,
    BaseMessage,
    HumanMessage,
    SystemMessage,
    get_buffer_string,
)
from pydantic.v1 import validator

from ark.core.idl.maas_protocol import MaasChatRequest
from ark.core.idl.ark_protocol import ArkChatRequest


class CustomPromptTemplate(BaseChatPromptTemplate):
    input_variables: List[str] = ["messages"]
    """List of input variables in template messages. Used for validation."""

    template: Template
    """
    Jinja2 template for prompt generation. Variables available in the context are:
    - time_info, string, formatted datetime. e.g., "2022年10月25日15时星期一".
    - location_info, string, if not provided use empty string. e.g., "北京市海淀区".
    - systems, list of string, system prompts.
    - questions, list of string, human prompts, including the current question.
    - answers, list of string, AI answers.
    - query, the current question.
    """

    keep_history_systems: bool = False
    keep_history_answers: bool = False
    keep_history_questions: bool = False
    chat_history_keep_human: bool = False
    chat_history_keep_ai: bool = False
    chat_history_len_limit: int = 0

    @validator("template", pre=True)
    def validate_template(cls, v: Union[str, Template]) -> Template:
        if isinstance(v, str):
            env = jinja2.Environment()

            def _datetime_format(
                    value: Union[str, datetime], format: Optional[str] = None
            ) -> str:
                def _gen_time_info(t: datetime) -> str:
                    weekdays = [
                        "星期一",
                        "星期二",
                        "星期三",
                        "星期四",
                        "星期五",
                        "星期六",
                        "星期日",
                    ]  # `tm_wday` 0~6, 6 means Sunday
                    return f'{t.strftime("%Y年%m月%d日%H时")}{weekdays[t.weekday()]}'

                if isinstance(value, str):
                    value = datetime.fromisoformat(value)

                if not format:
                    return _gen_time_info(value)

                return value.strftime(format)

            env.filters["datetime_format"] = _datetime_format
            return env.from_string(v)
        return v

    def _gen_location_info(
            self, location_info: Optional[Tuple[str, str]] = None
    ) -> str:
        city, district = location_info if location_info else ("", "")
        return city + district

    def _must_str(self, content: Union[str, List[Union[str, Dict]], Dict]) -> str:
        if isinstance(content, str):
            return content
        if isinstance(content, list):
            return "\n".join([self._must_str(c) for c in content])
        if isinstance(content, dict):
            return ""

    def _validate_and_fetch_questions_and_answers(
            self, messages: List[BaseMessage]
    ) -> Tuple[List[str], List[str], List[str]]:
        systems: List[str] = [
            self._must_str(msg.content) for msg in messages if msg.type == "system"
        ]
        questions: List[BaseMessage] = [
            msg for msg in messages if msg.type == "human"
        ]
        answers: List[BaseMessage] = [
            msg for msg in messages if msg.type == "ai"
        ]

        return (
            systems,
            [self._must_str(msg.content) for msg in questions],
            [self._must_str(msg.content) for msg in answers],
        )

    def _build_chat_history(self, messages: List[BaseMessage]) -> str:
        histories: List[Union[AIMessage, HumanMessage]] = []
        if len(messages) == 0:
            raise ValueError("No user question found in the request")
        elif len(messages) > 1:
            for msg in messages[:-1]:
                if isinstance(msg, HumanMessage) and self.chat_history_keep_human:
                    histories.append(msg)
                elif isinstance(msg, AIMessage) and self.chat_history_keep_ai:
                    histories.append(msg)

        if self.chat_history_len_limit > 0:
            history_len = sum([len(msg.content) for msg in histories])
            while history_len > self.chat_history_len_limit:
                history_len -= len(histories.pop(0).content)

        if len(histories) > 0:
            chat_history = get_buffer_string(
                histories, human_prefix="User", ai_prefix="Assistant"
            )
        else:
            chat_history = "无"
        return chat_history

    def format_messages(self, **kwargs: Any) -> List[BaseMessage]:
        """Format original messages into a single Human message.

        Args:
            messages: list of original messages.
            time_info: datetime, if not provided use current time.
            user_info: dict, some user info, e.g., {"city": "北京", "district": "海淀"}

        Returns:
            Formatted message
        """
        if "messages" not in kwargs:
            raise ValueError("Must provide messages: List[BaseMessage]")

        messages: List[BaseMessage] = kwargs.pop("messages")
        systems, questions, answers = self._validate_and_fetch_questions_and_answers(
            messages
        )
        time_info = kwargs.pop("time_info", datetime.now())

        user_info = kwargs.pop("user_info", {})
        city, district = (
            user_info.get("city", ""),
            user_info.get("district", ""),
        )
        location_info = self._gen_location_info(location_info=(city, district))
        chat_history = self._build_chat_history(messages)

        final_prompt = self.template.render(
            time_info=time_info,
            location_info=location_info,
            systems=systems,
            questions=questions,
            answers=answers,
            query=questions[-1] if len(questions) > 0 else "",
            user_info=user_info,
            chat_history=chat_history,
            **kwargs,  # for extra variables
        )

        if (
                not self.keep_history_systems
                and not self.keep_history_answers
                and not self.keep_history_questions
        ):
            return [HumanMessage(content=final_prompt)]
        elif (
                not self.keep_history_answers and not self.keep_history_questions
        ):  # keep system messages
            return [SystemMessage(content=system) for system in systems] + [
                HumanMessage(content=final_prompt)
            ]
        else:  # keep human or ai messages
            final_messages = []
            for message in messages:
                if message.type == "system" and not self.keep_history_systems:
                    continue
                elif message.type == "human":
                    if not self.keep_history_questions:
                        message = message.copy(update={"content": ""})
                elif message.type == "ai" and not self.keep_history_answers:
                    message = message.copy(update={"content": ""})
                final_messages.append(message)
            return final_messages


def format_current_meta_info(request: Union[MaasChatRequest, ArkChatRequest]) -> str:
    user_info = request.get_user_info_extra()
    meta_info = f"当前时间：{format_time_info(int(time.time()))}"
    if user_info:
        meta_info += f"\n当前位置：{user_info.city or ''}{user_info.district or ''}"
    return meta_info


def format_time_info(timestamp: int, zone: str = 'Asia/Shanghai') -> str:
    dt = datetime.fromtimestamp(timestamp)
    cst = pytz.timezone(zone)
    dt_cst = dt.astimezone(cst)
    weekdays = ["星期一", "星期二", "星期三", "星期四", "星期五", "星期六", "星期日"]
    _time_info = dt_cst.strftime("%Y年%m月%d日 %H:%M:%S")
    _time_info += "(CST) "
    _time_info += weekdays[dt_cst.weekday()]
    return _time_info



from typing import Any, List, Union

from jinja2 import Template
from langchain.prompts.chat import BaseChatPromptTemplate
from langchain.schema.messages import (
    AIMessage,
    BaseMessage,
    HumanMessage,
    get_buffer_string,
)
from pydantic.v1 import validator


class AbstractAugmentPromptTemplate(BaseChatPromptTemplate):
    """Abstract augment prompt template."""

    input_variables: List[str] = ["messages"]
    """List of input variables in template messages. Used for validation."""

    template: Template

    @validator("template", pre=True)
    @classmethod
    def validate_template(cls, v: Union[str, Template]) -> Template:
        if isinstance(v, str):
            return Template(source=v)
        return v

    @property
    def template_str(self) -> str:
        """Return the template string."""
        return self.template.render()

    def format_messages(self, **kwargs: Any) -> List[BaseMessage]:
        """Format original messages into a list of messages.
        Args:
            messages: list of original messages.
        Returns:
            Formatted message
        """
        if "reference" not in kwargs:
            raise ValueError("Must provide reference: str")
        reference = kwargs.pop("reference")
        if "last_abstract" not in kwargs:
            raise ValueError("Must provide last_abstract: str")
        last_abstract = kwargs.pop("last_abstract")

        abstract_user_prompt = self.template.render(
            last_abstract=last_abstract, reference=reference, **kwargs
        )
        return [HumanMessage(content=abstract_user_prompt, additional_kwargs=kwargs)]


class AugmentPromptTemplate(BaseChatPromptTemplate):
    """summary augment for chunk prompt template."""

    input_variables: List[str] = ["messages"]
    """List of input variables in template messages. Used for validation."""

    template: Template

    @validator("template", pre=True)
    @classmethod
    def validate_template(cls, v: Union[str, Template]) -> Template:
        if isinstance(v, str):
            return Template(source=v)
        return v

    @property
    def template_str(self) -> str:
        """Return the template string."""
        return self.template.render()

    def format_messages(self, **kwargs: Any) -> List[BaseMessage]:
        """Format original messages into a list of messages.
        Args:
            messages: list of original messages.
        Returns:
            Formatted message
        """
        if "reference" not in kwargs:
            raise ValueError("Must provide reference: str")
        reference = kwargs.pop("reference")

        prompt = self.template.render(reference=reference, **kwargs)
        return [HumanMessage(content=prompt, additional_kwargs=kwargs)]


class RagInnerPromptTemplate(BaseChatPromptTemplate):
    """prompt template for intent and query rewrite."""

    input_variables: List[str] = ["messages"]
    """List of input variables in template messages. Used for validation."""

    template: Template
    keep_human_history: bool = True
    keep_ai_history: bool = True
    max_history_len: int = 0

    @validator("template", pre=True)
    def validate_template(cls, v: Union[str, Template]) -> Template:
        if isinstance(v, str):
            return Template(source=v)
        return v

    @property
    def template_str(self) -> str:
        """Return the template string."""
        return self.template.render()

    def format_messages(self, **kwargs: Any) -> List[BaseMessage]:
        """Format original messages into a list of messages.

        Args:
            messages: list of original messages.

        Returns:
            Formatted message
        """
        if "messages" not in kwargs:
            raise ValueError("Must provide messages: List[BaseMessage]")
        messages = kwargs.pop("messages")
        histories: List[Union[AIMessage, HumanMessage]] = []
        if len(messages) == 0:
            raise ValueError("No user question found in the request")
        elif len(messages) > 1:
            current = messages[-1]
            for msg in messages[:-1]:
                if isinstance(msg, HumanMessage) and self.keep_human_history:
                    histories.append(msg)
                elif isinstance(msg, AIMessage) and self.keep_ai_history:
                    histories.append(msg)
        else:
            histories, current = [], messages[0]

        if self.max_history_len > 0:
            history_len = sum([len(msg.content) for msg in histories])
            while history_len > self.max_history_len:
                history_len -= len(histories.pop(0).content)

        if len(histories) > 0:
            chat_history = get_buffer_string(
                histories, human_prefix="User", ai_prefix="Assistant"
            )
        else:
            chat_history = "无"
        prompt = self.template.render(
            chat_history=chat_history, question=current.content, **kwargs
        )

        return [HumanMessage(content=prompt, additional_kwargs=kwargs)]


class IntentionPromptTemplate(RagInnerPromptTemplate):
    """Retriever intention template."""

    keep_human_history: bool = True
    keep_ai_history: bool = False
    max_history_len: int = 1000


class QueryRewritePromptTemplate(RagInnerPromptTemplate):
    """Retriever condense question prompt template."""

    keep_human_history: bool = True
    keep_ai_history: bool = True
    max_history_len: int = 2000


ABSTRACT_AUGMENT_TEMPLATE = Template(
    """# 角色
你是一个擅长总结摘要的专家，根据用户的参考资料和要求，生成总结摘要。

# 参考资料
## 新增资料
<资料开始>
{{reference}}
<资料结束>

## 前文总结
{{last_abstract}}

# 要求
1. 结合参考资料中的「前文总结」和「新增资料」，\
「前文总结」的信息必须完整保留，将「前文总结」和「新增资料」合并为完整的总结摘要；
2. 如果「前文总结」和「新增资料」识别到章节、分段等信息，则分章节进行汇总摘要；
3. 分段进行总结汇总，不能过于简略，保留参考资料的关键信息，包括但不限于：\
人名、地名、有名称的实体、时间、发生的事件、影响面、讨论的对象；
4. 可以直接引用重要的片段，重要的任务、对象在首次出现时必须使用完整的名称，\
注意引用次数不超过 5 次；
5. 直接输出总结摘要标题和正文，第一行是摘要的标题，第二行开始是摘要内容，\
无需添加任何开始和结束的标志，不要使用 markdown 格式，不要添加“正文摘要：”等前缀和说明

## 样例输出格式
《简述牛顿和天体力学》的摘要
艾萨克·牛顿（Isaac Newton）通过观察苹果落地的现象，深入思考并进行了一系列的实验和研究，\
最终发现了重力的存在和作用原理。
牛顿在他的著作《天体力学》，在这一领域他提出了行星运动的三定律，并利用万有引力定律解释了行星轨道的形状和运动规律：
第一定律（椭圆轨道定律）：行星绕太阳运动的轨道是一个椭圆，太阳位于椭圆的一个焦点上。
第二定律（面积速度定律）：行星在其轨道上的运动速度随其位置的变化而变化。当行星离太阳较近的时，\
它沿着轨道移动快；而当行星离太阳远时，它移动更慢，行星在相同时间内扫过的面积相等。
第三定律（调和定律）：行星轨道的周期与其轨道半长轴的关系成正比。行星离太阳越远，绕太阳运动所需的时间就越长。

# 任务
请根据上述要求和参考资料进行总结摘要
"""
)
abstract_augment_prompt_template = AbstractAugmentPromptTemplate(
    template=ABSTRACT_AUGMENT_TEMPLATE
)

SUMMARY_AUGMENT_TEMPLATE = Template(
    """# 角色
你是一位善于对「用户输入」进行重写的文本处理专家，在保证信息完整性的前提下，对文本进行总结重写。

# 约束
1. **列举主体时，必须保留「用户输入」中出现的全部对象，不得省略，且保留完整名称**；
2. 去除冗余信息，将类比、比喻、举例说明都精简为直接的形容词；
3. 使用平铺直叙的句式，将倒装句、反问句都转述为简短的直叙；
4. 直接输出符合要求的文本，不解释自己或者提供任何与摘要无关的信息，\
不要添加“以下是生成的摘要”等字句。
5. 如果「用户输入」无法提取有效的信息，例如「用户输入」是表格的片段、\
完全缺失语义信息的乱码，只输出"无法提取"；

# 任务
对用户输入执行信息提取，输出两行内容，第一行直接列举文本中出现的所有实体对象，\
第二行为总结重写文本，**确保信息的完整性**

## 样例输入输出
用户输入：
阿吉教你选择健康环保的水瓶，目前来看最安全的用玻璃，因为玻璃具有良好的热稳定性，\
最方便的用塑料，不易碎且轻便，但塑料有不同级别的安全性，最好使用安全性较高就选不含双酚A的PPSU塑料。

样例输出：
对象：阿吉、玻璃、塑料、双酚A、PPSU塑料
总结：选择水瓶时，要考虑安全性和便利性，玻璃和塑料各有优势。用玻璃最安全，\
但不如塑料轻便，高价的不含双酚A的PPSU塑料兼顾安全和便利。

# 用户输入
{{reference}}
"""
)
summary_augment_prompt_template = AugmentPromptTemplate(
    template=SUMMARY_AUGMENT_TEMPLATE
)

HYPO_QUERIES_AUGMENT_TEMPLATE = Template(
    """# 角色
你是一个文本处理专家，擅长对文本进行抽取和归纳出问题，最后输出成更简洁和清晰的问题，提出可以使用「用户输入」来回答的问题。

# 约束
- 先推理并替换掉「用户输入」中的指示代词，生成出来的问题要有清晰的主语，\
禁止使用指示代词；
- 补全或改写「用户输入」中的关键词，并保留到问题中，\
尽可能地包含「用户输入」中的关键词和实体词；
- 从不同角度提出「用户输入」主题有代表性的问题，\
问题可以从时间、地点、人物、事件、影响等多个角度抽取；
- 每个问题一行，不要添加序号和问题以外的任何内容，\
不要解释自己，不要重复；
- 如果「用户输入」无法提取有效的信息，\
例如「用户输入」是表格的片段、完全缺失语义信息的乱码，只输出"无法提取"；

# 任务
对用户输入执行信息提取，将「用户输入」分段按顺序抽取1~10个问题

## 样例输入输出
用户输入：
苹果会落在地上是因为地球对它施加了引力，也就是重力。重力是一种基本的物理现象，它导致物体受到朝向地球中心的吸引力。\
具体来说，地球的质量非常大，这使得它对周围的物体产生了引力。当苹果挂在树上时，树枝对苹果有一个支撑力，使其保持在树上。\
但是，当支撑力消失（例如树枝断裂或苹果自然脱落）时，苹果就会受到重力的作用而向下掉落。\
牛顿通过观察苹果落地的现象，深入思考并进行了一系列的实验和研究，最终发现了重力的存在和作用原理。

样例输出：
为什么苹果会落在地上，是什么导致了这种下落的物理现象？
重力为什么会导致苹果从树上掉落到地面？
牛顿是怎么从苹果落地发现重力的？

# 用户输入
{{reference}}
"""
)
hypo_queries_augment_prompt_template = AugmentPromptTemplate(
    template=HYPO_QUERIES_AUGMENT_TEMPLATE
)

FAQ_AUGMENT_TEMPLATE = Template(
    """# 角色
你是一个文本处理专家，擅长对文本进行抽取和归纳出高频问题和回答，根据「用户输入」构造出问答对话。

# 约束
- 先推理并替换掉「用户输入」中的指示代词，\
生成出来的问题要有清晰的主语，禁止使用指示代词。
- 补全或改写「用户输入」中的关键词，并保留到问题中，\
尽可能地包含「用户输入」中的关键词和实体词。
- 从不同角度提出「用户输入」主题有代表性的问题，多个角度提问和回答，\
例如：「用户输入」是一条新闻，可以从时间、地点、人物、事件、影响面等角度提问；\
「用户输入」是一段说明书，可以从产品特性、功能、故障现象等角度提问。
- 每个问题一行，不要添加序号和问答对以外的任何内容，不要解释自己，不要重复。
- 如果「用户输入」无法提取有效的信息，\
例如「用户输入」是表格的片段、完全缺失语义信息的乱码，只输出一次"无法提取"。

# 任务
对用户输入执行信息提取，将「用户输入」分段按抽取成多个问答对，问题和答案各占一行，每个问答对之间空一行。
若能提取，第一个问答必须是针对「用户输入」全文的提问和回答，之后的问答对针对分段细节提问和回答。

## 样例输入输出
用户输入：
苹果会落在地上是因为地球对它施加了引力，也就是重力。\
重力是一种基本的物理现象，它导致物体受到朝向地球中心的吸引力。\
具体来说，地球的质量非常大，这使得它对周围的物体产生了引力。\
当苹果挂在树上时，树枝对苹果有一个支撑力，使其保持在树上。\
但是，当支撑力消失（例如树枝断裂或苹果自然脱落）时，\
苹果就会受到重力的作用而向下掉落。牛顿通过观察苹果落地的现象，\
深入思考并进行了一系列的实验和研究，最终发现了重力的存在和作用原理。

样例输出：
为什么苹果会落在地上，是什么导致了这种下落的物理现象？
苹果落地是因为地球的重力产生的物理现象，重力对物体产生朝向地球中心的吸引力。

重力为什么会导致苹果从树上掉落到地面？
重力本身不直接导致苹果掉落地面，当苹果挂在树上时，树枝对苹果有一个支撑力，使其保持在树上。\
但是，当支撑力消失（例如树枝断裂）时，重力大于支撑力，使得苹果向下掉落。

牛顿是怎么从苹果落地发现重力的？
牛顿通过观察察苹果落地，深入思考并进行实验和研究，最终发现了重力的存在和作用原理。

# 用户输入
{{reference}}
"""
)
faq_augment_prompt_template = AugmentPromptTemplate(template=FAQ_AUGMENT_TEMPLATE)

CONDENSE_QUESTION_PROMPT_TEMPLATE = Template(
    """# 角色
你是一个检索专家，擅长根据「聊天记录」和「新问题」，进行信息提取并将问题重写为适用于检索的新问题。

# 约束
- 先推理并替换掉「新问题」中的指示代词，生成出来的问题要有清晰的主语，禁止使用指示代词；
- 补全或改写「新问题」中的关键词，并保留到问题中，\
尽可能地包含「聊天记录」和「新问题」关联的关键词和实体词；
- 不要添加序号和问题以外的任何内容，不要解释自己，不要重复；

# 聊天记录
{{chat_history}}

# 新问题: 
{{question}}

# 任务
结合「聊天记录」和「新问题」，将新问题重新表述为单个完整独立的问题，直接输出重新表述的问题，不要添加根据聊天记录等话语。
"""
)
condense_question_prompt_template = QueryRewritePromptTemplate(
    template=CONDENSE_QUESTION_PROMPT_TEMPLATE
)

HYPO_ANSWER_PROMPT_TEMPLATE = Template(
    """# 角色
你是一个百科专家，擅长根据「聊天记录」和一个「新问题」，生成一段相关资料作为回答，可以忽略答案的事实性。

# 约束
- 先推理并替换掉「新问题」中的指示代词，生成出来的语料要有清晰的主语，禁止使用指示代词；
- 补全或改写「新问题」中的关键词，尽可能地包含「新问题」中的关键词和实体词；
- 不要添加序号和相关资料以外的任何内容，不要解释自己，不要重复；

# 聊天记录
{{chat_history}}

# 问题
{{question}}

# 任务
生成一段相关资料作为「新问题」的回答。
"""
)
hypo_answer_prompt_template = QueryRewritePromptTemplate(
    template=HYPO_ANSWER_PROMPT_TEMPLATE
)

KNOWLEDGE_INTENTION_TEMPLATE = Template(
    """请你根据历史对话记录和用户问题，判断你是否需要检索信息，回答用户问题。

请你以如下的格式回答：
回答只有一行，只能是"无需检索"或者"需要检索"，不要再进行任何补充描述。

历史问题记录:
{{chat_history}}

用户问题:
{{question}}

回答:
"""
)

ABSTRACT_INTENTION_TEMPLATE = Template(
    """请判断回答用户问题是否要用到参考资料的摘要信息?

示例:
xx论文的摘要是什么?->是
xx的主要内容是什么?->是
xx的yy章节主要内容是?->否
xx论文中在zz情况下应该怎么做?->否

请你以如下的格式回答：
回答只有一行，只能是"无需检索"或者"需要检索"，不要再进行任何补充描述。

用户问题:
{{question}}

回答:
"""
)

DEFAULT_SUMMARY_PROMPT = Template(
    """# 参考资料
{{reference}}

# 任务和限制
你需要根据参考资料来回答用户的问题，解决用户关于百科问题的疑问，你的回答需要准确和完整。

你的回答要满足以下几个限制：
1. 回答内容必须在参考资料范围内，不能做任何参考资料以外的扩展解释。
2. 如果参考资料不能帮助你回答用户问题，请回答“抱歉，这个问题我还不会”。
3. 你的回答不能暴露参考资料的存在，回答内容不能包含诸如“根据提供的参考资料”，\
“根据我的知识库”等，直接回答跟用户问题有关的内容即可。

# 任务执行
现在请你根据提供的参考资料，遵循限制来回答用户的问题，你的回答需要准确和完整。

用户问题：
{{question}}

完整、详细、有帮助的回答：
"""
)

RAG_INTENTION_TEMPLATE = Template(
    """# 任务描述
请根据「历史记录」和「用户问题」，判断应该以下哪些内容源哪些可以获得有助于回答用户问题的资料：
{{sources_desc_text}}

# 任务要求
1. 可以选择{{enum_list_text}}，或者选择「无需检索」，不要再进行任何补充描述；
2. 如果是闲聊或者是内容源无关的用户问题等情景，不需要额外信息帮助回答问题，只输出「无需检索」；

# 执行任务
历史记录：
{{chat_history}}

用户问题：
{{question}}

判断结果：
"""
)

# -*- Query Rewrite -*- #
RAG_REWRITE_PROMPT_TEMPLATE = Template(
    """# 聊天记录
{{chat_history}}
<聊天记录结束>

# 用户提问
{{question}}

# 角色
你是一个检索专家，擅长根据「聊天记录」和「用户提问」，进行信息提取并将问题重写为适用于检索的1~3个问题。

# 约束
- 先推理并替换掉「用户提问」中的指示代词，生成出来的问题要有清晰的主语，禁止使用指示代词；
- 补全或改写「用户提问」中的关键词，并保留到问题中，尽可能地包含「聊天记录」和「用户提问」关联的关键词和实体词；
- 直接输出重新表述的问题，不要添加根据聊天记录等话语，不要添加序号和问题以外的任何内容，不要解释自己，不要重复；

# 任务
结合「聊天记录」和「用户提问」，将用户提问重新表述为若干个完整且独立的问题，每行输出一个改写的问题，每个问题不超过100字；
第一个改写的问题保持和「用户提问」的拥有一致的意思，从第二个改写问题开始，提出有助于检索相关资料的问题。

改写的问题：
"""
)

# -*- Summary Generation -*- #
RAG_SUMMARY_TEMPLATE = Template(
    """# 历史对话
{{chat_history}}

# 参考资料
{{reference}}

# 当前环境信息
{{meta_info}}

# 参考资料
{{reference}}

# 任务
{{task_description}}

# 任务执行
现在请你根据提供的参考资料，遵循限制来回答用户的问题，你的回答需要准确和完整；

用户问题：
{{question}}

你的回答：
"""
)

RAG_STRICT_TASK_DESCRIPTION = """- 你需要根据「参考资料」和「当前环境信息」完整、详细、有帮助的回答「用户问题」。
- 你的回答要满足以下几个限制：
    1. 回答内容必须在参考资料范围内，不能做任何参考资料以外的扩展解释；
    2. 如果参考资料不能帮助你回答用户问题，请回答“抱歉，这个问题我还不会，尝试告诉我更多信息吧”；
    3. 你的回答不能暴露参考资料的存在，回答内容不能包含诸如“根据提供的参考资料”，“根据我的知识库”等，直接回答跟用户问题有关的内容即可；
"""

RAG_PARTIAL_TASK_DESCRIPTION = """
- 你需要根据「参考资料」和「当前环境信息」完整、详细、有帮助的回答「用户问题」；
- 回答问题时，先查阅与「用户问题」有关联的「参考资料」、「当前环境信息」，结合你自己的知识，汇总信息后，给出对「用户问题」有帮助的回答；
- 你的回答要满足以下几个限制：
    1. 使用专业的语言风格回答，你的知识和参考资料不一致时，以参考资料为准；
    2. 回答内容不能包含诸如“根据提供的参考资料”，“根据我的知识库”等，直接回答跟用户问题有关的内容即可；
"""



import copy
from enum import Enum
from typing import Any, Dict, List, Optional, Union

from pydantic import BaseModel, Field

from ark.core.idl.ark_protocol import (
    ArkChatParameters,
    ArkMessage,
)
from ark.core.idl.common_protocol import ModelDesc

RAG_SOURCE = ["rag"]
RAG_TAG = {"rag": "知识库"}

RAG_DESC = {"rag": ""}

SOURCE_PRIPORITY = [
    "weather",
    "douyin_baike",
    "toutiao_article",
    "douyin_short_video",
    "search_engine",
    "xigua_feed_video",
    "toutiao_short_content",
]


class Prompt(BaseModel):
    template: str = ""


class PhaseMode(str, Enum):
    SKIP = "Skip"
    CUSTOM = "Custom"
    STRICT = "Strict"
    PARTIAL = "Partial"


class EndpointConfig(BaseModel):
    model_desc: ModelDesc = Field(default_factory=ModelDesc)
    parameters: Optional[ArkChatParameters] = Field(default_factory=ArkChatParameters)
    """
    pre-defined parameters
    """
    system_prompts: Optional[List[str]] = None
    """
    pre-defined sp
    """
    mode: PhaseMode = PhaseMode.CUSTOM
    """
    model calling mode
    """
    prompt: Optional[Prompt] = None
    """
    prompt template for model
    """
    custom_variables: Dict[str, Any] = Field(default_factory=dict)
    """
    custom_variables for prompt template variables, 
    keys & values stand for variable name & value
    """
    max_history_length: int = 1500
    max_reference_token: int = 3300
    min_reference_token: int = 200
    max_keywords_length: int = 3

    additional_configs: Dict[str, "EndpointConfig"] = Field(default_factory=dict)

    def get_model_ep(self) -> str:
        return self.model_desc.endpoint_id or ""

    def merge_messages(self, messages: List[ArkMessage]) -> Optional[List[ArkMessage]]:
        if not self.system_prompts:
            return messages
        else:
            prompts = [
                ArkMessage(role="system", content=system_prompt)
                for system_prompt in self.system_prompts
            ]
            prompts.extend(messages)
            return prompts

    def merge_parameters(
        self, parameters: Optional[Union[ArkChatParameters, Dict[str, Any]]]
    ) -> Optional[Union[ArkChatParameters, Dict[str, Any]]]:
        if not parameters:
            return self.parameters
        elif not self.parameters:
            return parameters
        else:
            return self.parameters.merge_from(parameters)

    # merge endpoint config
    def merge_from(self, other: Optional["EndpointConfig"] = None) -> "EndpointConfig":
        other_dict = other.__dict__ if other else {}
        merged_dict = copy.deepcopy(self.__dict__)

        for key in merged_dict.keys():
            value = other_dict.get(key, None)
            if value is not None:
                if isinstance(value, ModelDesc):
                    merged_dict[key] = ModelDesc(
                        name=value.name or merged_dict[key].name,
                        version=value.version or merged_dict[key].version,
                        endpoint_id=value.endpoint_id or merged_dict[key].endpoint_id,
                        authorization_token=value.authorization_token
                        or merged_dict[key].authorization_token,
                    )
                elif value:
                    merged_dict[key] = value

        return self.__class__(**merged_dict)


def default_endpoint_config(endpoint_id: str) -> EndpointConfig:
    return EndpointConfig(model_desc=ModelDesc(endpoint_id=endpoint_id))



from abc import ABC
from typing import Any, Dict, List, Optional, Union

from ark.component.llm.base import BaseEmbeddingLanguageModel
from ark.component.retriever.base_retriever import BaseRetriever
from ark.component.vectorstores.embedding import norm_l2
from ark.component.vectorstores.vikingdb import VikingDB, VikingDBSchema
from ark.core.rag import KnowledgeChunk


class VikingDBRetriever(BaseRetriever, ABC):
    def __init__(self, config: VikingDBSchema, db: VikingDB) -> None:
        self.config = config
        self.db = db
        self._embedding_cache: Dict[str, Any] = {}

    async def get_embedding(self, query: str) -> List[float]:
        if query in self._embedding_cache:
            return self._embedding_cache[query]
        llm = BaseEmbeddingLanguageModel(
            endpoint_id=self.config.vector.get("model") or "", input=[query]
        )
        resp = await llm.arun()
        assert resp.data and len(resp.data) > 0, f"invalid embedding response: {resp}"
        self._embedding_cache[query] = resp.data[0].embedding
        return resp.data[0].embedding

    async def retrieve(
        self,
        query: Union[str, List[float]],
        retrieve_count: int,
        dsl_filter: Optional[Any] = None,
        partition: str = "default",
        norm: str = "none",
        vector_dim: int = 0,
        instruct: str = "为这个句子生成表示以用于检索相关文章：",
    ) -> List[KnowledgeChunk]:
        if dsl_filter:
            db_filter = dsl_filter
        else:
            db_filter = None
        if isinstance(query, str) and self.config.vector.get("embedding_type") == "llm":
            sliced_emb = await self.get_embedding(instruct + query)
            if vector_dim > 0 and len(sliced_emb) != vector_dim:
                # need vector slice & norm
                sliced_emb = sliced_emb[:vector_dim]
            if norm == "l2":
                emb_query = norm_l2(sliced_emb)
            else:
                emb_query = sliced_emb
            return self.db.search(
                query=emb_query,
                count=retrieve_count,
                filter=db_filter,
                partition=partition,
            )
        return self.db.search(
            query=query, count=retrieve_count, filter=db_filter, partition=partition
        )

    async def chunk_diffusion(
        self,
        query: Union[str, List[Any]],
        chunk: KnowledgeChunk,
        partition: str = "default",
        match_scalars: List[str] = ["chunk_type", "doc_id"],
        forward_diffusion: int = 1,
        backward_diffusion: int = 1,
    ) -> List[KnowledgeChunk]:
        retrieve_count = forward_diffusion + backward_diffusion + 1
        conds = [
            {
                "op": "must",
                "field": field_name,
                "conds": [chunk.scalars[field_name]],
            }
            for field_name in match_scalars
        ]
        conds.append(
            {
                "op": "range",
                "field": "chunk_id",
                "gte": max(chunk.scalars["chunk_id"] - backward_diffusion, 0),
                "lt": chunk.scalars["chunk_id"] + forward_diffusion + 1,
            }
        )
        db_filter = {"op": "and", "conds": conds}
        if isinstance(query, str) and self.config.vector.get("embedding_type") == "llm":
            emb_query = await self.get_embedding(query)
            return self.db.search(
                query=emb_query,
                count=retrieve_count,
                filter=db_filter,
                partition=partition,
            )
        return self.db.search(
            query=query, count=retrieve_count, filter=db_filter, partition=partition
        )



import abc
import json
import logging
from typing import Any, Callable, Iterable, List, TypeVar

from pydantic import BaseModel

from ark.core.idl.common_protocol import Request, Response
from ark.core.idl.maas_protocol import MaasChatChoice, MaasChatRequest, MaasChatResponse
from ark.core.task import task
from ark.core.utils.context import get_reqid
from ark.core.utils.errors import GPTException, InternalServiceError

RequestType = TypeVar("RequestType", bound=Request)
ResponseType = TypeVar("ResponseType", bound=Response)


class SyncRunner(BaseModel):
    invoke: Callable[[RequestType], Iterable[ResponseType]]

    class Config:
        """Configuration for this pydantic object."""

        arbitrary_types_allowed = True

    def __init__(
        self,
        runnable_func: Callable[[RequestType], Iterable[ResponseType]],
        **kwargs: Any,
    ):
        super().__init__(invoke=runnable_func, **kwargs)

    @abc.abstractmethod
    def run(self, request: RequestType) -> Response:
        pass

    @abc.abstractmethod
    def generate(self, request: RequestType) -> Iterable[str]:
        pass


class ChatSyncRunner(SyncRunner):
    def __init__(
        self,
        runnable_func: Callable[[MaasChatRequest], Iterable[MaasChatResponse]],
        **kwargs: Any,
    ):
        super().__init__(runnable_func=runnable_func, **kwargs)

    @task(distributed=False)
    def run(self, request: MaasChatRequest) -> MaasChatResponse:
        verbose: List[List[MaasChatChoice]] = []
        responses: List[MaasChatResponse] = []

        try:
            for resp in self.invoke(request):  # type: MaasChatResponse
                if resp.choices and resp.choices[0].message:
                    responses.append(resp)
                else:
                    verbose.append(resp.choices)
        except GPTException as e:
            resp = MaasChatResponse(req_id=get_reqid(), error=e.to_error())
            responses.append(resp)
            logging.exception("assistant chat meet gpt error")
        except Exception as e:
            err = InternalServiceError(str(e))
            resp = MaasChatResponse(req_id=get_reqid(), error=err.to_error())
            responses.append(resp)
            logging.exception("assistant chat meet internal error")

        response = MaasChatResponse.merge(responses=responses)
        return response

    @task(distributed=False)
    def generate(self, request: MaasChatRequest) -> Iterable[str]:
        try:
            for resp in self.invoke(request):  # type: MaasChatResponse
                if resp.choices and resp.choices[0].message:
                    if isinstance(resp, BaseModel):
                        yield f"data:{json.dumps(resp.model_dump(exclude_unset=True), ensure_ascii=False)}\r\n\r\n"  # noqa E501
                    else:
                        yield f"data:{resp}\r\n\r\n"  # noqa E501
        except GPTException as e:
            resp = MaasChatResponse(req_id=get_reqid(), error=e.to_error())
            logging.exception("stream chat meet error")
            yield f"data:{json.dumps(resp.model_dump(exclude_unset=True), ensure_ascii=False)}\r\n\r\n"  # noqa E501
        except Exception as e:
            err = InternalServiceError(str(e))
            resp = MaasChatResponse(req_id=get_reqid(), error=err.to_error())
            logging.exception("stream chat meet error")
            yield f"data:{json.dumps(resp.model_dump(exclude_unset=True), ensure_ascii=False)}\r\n\r\n"  # noqa E501
        yield "data:[DONE]\r\n\r\n"



import logging
from typing import Any

from .common import LoggerName, Timer

__all__ = ["DEBUG", "INFO", "WARN", "ERROR", "Timer"]


def DEBUG(msg: str, *args: Any, **kwargs: Any) -> None:
    logging.getLogger(LoggerName.get()).debug(msg, stacklevel=2, *args, **kwargs)


def INFO(msg: str, *args: Any, **kwargs: Any) -> None:
    logging.getLogger(LoggerName.get()).info(msg, stacklevel=2, *args, **kwargs)


def WARN(msg: str, *args: Any, **kwargs: Any) -> None:
    logging.getLogger(LoggerName.get()).warn(msg, stacklevel=2, *args, **kwargs)


def ERROR(msg: str, *args: Any, **kwargs: Any) -> None:
    logging.getLogger(LoggerName.get()).error(msg, stacklevel=2, *args, **kwargs)



from typing import AsyncIterator, Optional

from ark.component.llm.utils import get_maas_client_v2
from ark.core._api.deprecation import deprecated
from ark.core.idl.maas_protocol import (
    MaasChatRequest,
    MaasChatResponse,
    MaasClassificationRequest,
    MaasClassificationResponse,
    MaasEmbeddingsRequest,
    MaasEmbeddingsResponse,
    MaasTokenizeRequest,
    MaasTokenizeResponse,
)
from ark.core.task import task
from ark.core.utils.types import dump_json


@deprecated(
    since="0.1.11",
    removal="0.2.0",
    alternative_import="ark.component.v3.llm.BaseEmbeddingLanguageModel",
)
@task()
async def embeddings(
    endpoint_id: str,
    request: MaasEmbeddingsRequest,
    host: str = "maas-api.ml-platform-cn-beijing.volces.com",
    region: str = "cn-beijing",
    ak: Optional[str] = None,
    sk: Optional[str] = None,
) -> MaasEmbeddingsResponse:
    maas = get_maas_client_v2(host, region, ak, sk)
    return await maas.async_embeddings(endpoint_id, dump_json(request))


@deprecated(
    since="0.1.11",
    removal="0.2.0",
    alternative_import="ark.component.v3.llm.BaseChatLanguageModel",
)
@task()
async def chat(
    endpoint_id: str,
    request: MaasChatRequest,
    host: str = "maas-api.ml-platform-cn-beijing.volces.com",
    region: str = "cn-beijing",
    ak: Optional[str] = None,
    sk: Optional[str] = None,
) -> MaasChatResponse:
    maas = get_maas_client_v2(host, region, ak, sk)
    req = dump_json(request)
    return await maas.async_chat(endpoint_id, req)


@deprecated(
    since="0.1.11",
    removal="0.2.0",
    alternative_import="ark.component.v3.llm.BaseChatLanguageModel",
)
@task()
async def streaming_chat(
    endpoint_id: str,
    request: MaasChatRequest,
    host: str = "maas-api.ml-platform-cn-beijing.volces.com",
    region: str = "cn-beijing",
    ak: Optional[str] = None,
    sk: Optional[str] = None,
) -> AsyncIterator[MaasChatResponse]:
    maas = get_maas_client_v2(host, region, ak, sk)
    return await maas.async_stream_chat(endpoint_id, dump_json(request))


@deprecated(
    since="0.1.11",
    removal="0.2.0",
    alternative_import="ark.component.v3.llm.BaseTokenizeLanguageModel",
)
@task()
async def tokenize(
    endpoint_id: str,
    request: MaasTokenizeRequest,
    host: str = "maas-api.ml-platform-cn-beijing.volces.com",
    region: str = "cn-beijing",
    ak: Optional[str] = None,
    sk: Optional[str] = None,
) -> MaasTokenizeResponse:
    maas = get_maas_client_v2(host, region, ak, sk)
    return await maas.async_tokenize(endpoint_id, dump_json(request))


@deprecated(
    since="0.1.11",
    removal="0.2.0",
    alternative_import="ark.component.v3.llm.BaseClassificationLanguageModel",
)
@task()
async def classification(
    endpoint_id: str,
    request: MaasClassificationRequest,
    host: str = "maas-api.ml-platform-cn-beijing.volces.com",
    region: str = "cn-beijing",
    ak: Optional[str] = None,
    sk: Optional[str] = None,
) -> MaasClassificationResponse:
    maas = get_maas_client_v2(host, region, ak, sk)
    return await maas.async_classification(endpoint_id, dump_json(request))



from typing import Any, Dict, Literal

from pydantic import Field

from ark.component.prompts.group_chat_prompt import GroupChatPromptTemplate
from ark.component.v3.llm.base import BaseChatLanguageModel, BaseChatPromptTemplate
from ark.core.idl.ark_protocol import ArkCharacterConfig, ArkGroupChatConfig
from ark.core.utils.errorsv3 import InvalidParameter, MissingParameter


class GroupChat(BaseChatLanguageModel):
    name: Literal["GroupChat"] = "GroupChat"
    group_chat_config: ArkGroupChatConfig
    template: BaseChatPromptTemplate = Field(default_factory=GroupChatPromptTemplate)
    character_config_map: Dict[str, ArkCharacterConfig]

    def __init__(self, group_chat_config: ArkGroupChatConfig, **kwargs: Any):
        super().__init__(
            group_chat_config=group_chat_config,  # type: ignore
            character_config_map={c.name: c for c in group_chat_config.characters},  # type: ignore
            template=GroupChatPromptTemplate(group_chat_config=group_chat_config),
            **kwargs,
        )

    def get_request_model(self, **kwargs: Any) -> str:
        if "target_character_name" not in kwargs:
            raise MissingParameter(parameter="target_character_name")

        character_config = self.character_config_map.get(
            kwargs.get("target_character_name", "")
        )

        if not character_config:
            raise InvalidParameter(
                parameter="target_character_name",
                cause="specific character config not found",
            )

        if (
            not character_config.model_desc
            or not character_config.model_desc.endpoint_id
        ):
            return self.endpoint_id

        return character_config.model_desc.endpoint_id



import random
from datetime import datetime

time_fmt = "%Y%m%d%H%M%S"


def gen_log_id() -> str:
    return datetime.now().strftime(time_fmt) + format(
        random.randint(0, 2**64 - 1), "020X"
    )



from typing import Optional

from ark.core.client import ArkClient


def get_maas_client_v2(
    host: str = "maas-api.ml-platform-cn-beijing.volces.com",
    region: str = "cn-beijing",
    ak: Optional[str] = None,
    sk: Optional[str] = None,
) -> ArkClient:
    return ArkClient.get_instance_sync(host=host, region=region, ak=ak, sk=sk)



from typing import AsyncIterator

from aiohttp import StreamReader


class AsyncSSEDecoder(object):
    def __init__(self, source: StreamReader) -> None:
        self.source = source

    async def _read(self) -> AsyncIterator[bytes]:
        data = b""
        async for chunk in self.source:
            for line in chunk.splitlines(True):
                data += line
                if data.endswith((b"\r\r", b"\n\n", b"\r\n\r\n")):
                    yield data
                    data = b""
        if data:
            yield data

    async def next(self) -> AsyncIterator[bytes]:
        async for chunk in self._read():
            for line in chunk.splitlines():
                # skip comment
                if line.startswith(b":"):
                    continue

                if b":" in line:
                    field, value = line.split(b":", 1)
                else:
                    field, value = line, b""

                if field == b"data" and len(value) > 0:
                    yield value



from typing import Optional

from pydantic import BaseModel, Field

from ark.core._api.deprecation import deprecated
from ark.core.client import Client, get_client_pool
from ark.core.client.maas import ArkClient
from ark.core.idl.maas_protocol import SearchRequest, SearchResponse


def _get_ark_client() -> Optional[Client]:
    client_pool = get_client_pool()
    client = client_pool.get_client("chat")
    if not client:
        client = ArkClient()
    return client


@deprecated(
    since="0.1.11", removal="0.2.0", alternative_import="ark.component.v3.search.Search"
)
class Search(BaseModel):
    endpoint_id: str
    client: ArkClient = Field(default_factory=_get_ark_client)

    class Config:
        arbitrary_types_allowed = True

    async def arun(self, request: SearchRequest) -> SearchResponse:
        return await self.client.async_search(self.endpoint_id, request)
